'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Strategy augmented from graphql-subscriptions
 * https://github.com/graphql/graphql-js/blob/master
 * /src/subscription/__tests__/eventEmitterAsyncIterator.js
 *
 * Added support for debounce
 */

// from iterall https://github.com/leebyron/iterall/
var SYMBOL_ASYNC_ITERATOR = typeof Symbol === 'function' && Symbol.asyncIterator;
var $$asyncIterator = SYMBOL_ASYNC_ITERATOR || '@@asyncIterator';

function isObject(obj) {
  return (typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' && obj !== null && !Array.isArray(obj);
}

function call(func) {
  if (typeof func === 'function') {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    return func.apply(undefined, args);
  }
}

var AsyncIteratorBacking = function () {
  function AsyncIteratorBacking(values, options) {
    var _this2 = this;

    _classCallCheck(this, AsyncIteratorBacking);

    var _values = values;
    var _options = options;

    if (isObject(values)) {
      _options = values;
      _values = [];
    }

    var _Object$assign = Object.assign({}, _options),
        debounce = _Object$assign.debounce,
        onThrow = _Object$assign.onThrow,
        onReturn = _Object$assign.onReturn,
        onNext = _Object$assign.onNext,
        onPush = _Object$assign.onPush,
        onEnd = _Object$assign.onEnd;

    this.done = false;
    this.onThrow = onThrow;
    this.onReturn = onReturn;
    this.onPush = onPush;
    this.onEnd = onEnd;
    this.onNext = onNext;
    this.timeout = null;
    this.dataQueue = Array.isArray(_values) ? _values.slice() : [];
    this.awaitQueue = [];
    this.debounce = typeof debounce === 'number' && debounce > 0 ? debounce : null;

    // modify the values array to push new data. this way when
    // data is added to the array it can be queued
    if (Array.isArray(values)) {
      values.length = 0;
      values.push = function (value) {
        return _this2._push(value);
      };
    }
  }

  AsyncIteratorBacking.prototype.iterator = function iterator() {
    var _this = this;
    // build the iterator
    var iterator = _defineProperty({
      next: function next() {
        return _this.done && !_this.dataQueue.length ? iterator.return() : _this._next();
      },
      return: function _return() {
        call(this.onReturn);
        _this._end();
        return Promise.resolve({ value: undefined, done: true });
      },
      throw: function _throw(error) {
        call(this.onThrow, error);
        _this._end();
        return Promise.reject(error);
      }
    }, $$asyncIterator, function () {
      return iterator;
    });

    // return the pushable iterator object
    return iterator;
  };

  /**
   * Handles a push request with optional debounce
   * @param {*} value
   */


  AsyncIteratorBacking.prototype._push = function _push(value) {
    var _this3 = this;

    if (!this.debounce) {
      return this._pushValue(value);
    }

    // if using a debounce, clear the timeout
    // and then set a new one that handles the value
    clearTimeout(this.timeout);
    this.timeout = setTimeout(function () {
      _this3._pushValue(value);
    }, this.debounce);
  };

  /**
   * Pushes a value into the queue
   * @param {*} value
   */


  AsyncIteratorBacking.prototype._pushValue = function _pushValue(value) {
    call(this.onPush, value);
    return this.awaitQueue.length ? this.awaitQueue.shift()({ value: value, done: false }) : this.dataQueue.push(value);
  };

  /**
   * Gets the next value in the queue as a promise or
   * queues an await resolve
   */


  AsyncIteratorBacking.prototype._next = function _next() {
    var _this4 = this;

    call(this.onNext);
    return new Promise(function (resolve) {
      if (_this4.dataQueue.length) {
        return resolve({
          value: _this4.dataQueue.shift(),
          done: false
        });
      } else if (_this4.done) {
        return resolve({
          value: undefined,
          done: true
        });
      }
      _this4.awaitQueue.push(resolve);
    });
  };

  /**
   * Ends the iterator
   */


  AsyncIteratorBacking.prototype._end = function _end() {
    if (!this.done) {
      this.done = true;
      call(this.onEnd);
      call(this.cleanup);
      this.awaitQueue.forEach(function (resolve) {
        resolve({ value: undefined, done: true });
      });
      this.awaitQueue.length = 0;
      this.dataQueue.length = 0;
    }
  };

  return AsyncIteratorBacking;
}();

var AsyncIterator = exports.AsyncIterator = function AsyncIterator() {
  _classCallCheck(this, AsyncIterator);

  for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    args[_key2] = arguments[_key2];
  }

  return new (Function.prototype.bind.apply(AsyncIteratorBacking, [null].concat(args)))().iterator();
};

/**
* Default, create iterator from an array
* @param {*} array
*/


var fromArray = exports.fromArray = function fromArray(array, options) {
  return new AsyncIterator(array, options);
};

/**
* Create an iterator from a nodejs stream
* @param {*} stream
* @param {*} options
*/
var fromStream = exports.fromStream = function fromStream(stream, options) {
  var array = [];
  var opts = Object.assign({}, options);
  var DATA = opts.dataEvent || 'data';
  var CLOSE = opts.closeEvent || 'close';
  var backing = new AsyncIteratorBacking(array, opts);
  stream.on(DATA, function (value) {
    return array.push(value);
  });
  stream.on(CLOSE, function () {
    backing.done = true;
  });
  return backing.iterator();
};

/**
* Create an iterator from an event emitter
* @param {*} emitter
* @param {*} event
* @param {*} options
*/
var fromEvent = exports.fromEvent = function fromEvent(emitter, event, options) {
  var array = [];
  var opts = Object.assign({}, options);
  var backing = new AsyncIteratorBacking(array, opts);
  var closeEvent = opts.closeEvent || null;
  if (closeEvent) {
    emitter.on(closeEvent, function () {
      backing.done = true;
    });
  }
  backing.cleanup = function () {
    emitter.removeAllListeners(event);
    if (closeEvent) {
      emitter.removeAllListeners(closeEvent);
    }
  };
  emitter.on(event, function (value) {
    return array.push(value);
  });
  return backing.iterator();
};

/**
 * Export the backing so that it can be extended
 */
AsyncIterator.Backing = AsyncIteratorBacking;

AsyncIterator.fromArray = fromArray;
AsyncIterator.fromStream = fromStream;
AsyncIterator.fromEvent = fromEvent;